# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb.exceptions import VerificationException


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService,
         zset: bytes, seekKey: bytes, seekScore: float,
         seekAtTx: int, inclusive: bool, limit: int, desc: bool, minscore: float,
         maxscore: float, sinceTx, nowait):
    request = schema_pb2.ZScanRequest(
        set=zset,
        seekKey=seekKey,
        seekScore=seekScore,
        seekAtTx=seekAtTx,
        inclusiveSeek=inclusive,
        limit=limit,
        desc=desc,
        minScore=schema_pb2.Score(score=minscore),
        maxScore=schema_pb2.Score(score=maxscore),
        sinceTx=sinceTx,
        noWait=nowait,
    )
    msg = service.ZScan(request)
    return msg



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2

from immudb import header_manipulator_client_interceptor
from immudb.handler import (batchGet, batchSet, changePassword, changePermission, createUser,
                            currentRoot, databaseCreate, databaseList, databaseUse,
                            get, listUsers, sqldescribe, verifiedGet, verifiedSet, setValue, history,
                            scan, reference, verifiedreference, zadd, verifiedzadd,
                            zscan, healthcheck, txbyid, verifiedtxbyid, sqlexec, sqlquery,
                            listtables, execAll)
from immudb.rootService import *
from immudb.grpc import schema_pb2_grpc
import warnings
import ecdsa


class ImmudbClient:

    def __init__(self, immudUrl=None, rs: RootService = None, publicKeyFile: str = None):
        if immudUrl is None:
            immudUrl = "localhost:3322"
        self.channel = grpc.insecure_channel(immudUrl)
        self.__stub = schema_pb2_grpc.ImmuServiceStub(self.channel)
        if rs is None:
            self.__rs = RootService()
        else:
            self.__rs = rs
        self.__url = immudUrl
        self.loadKey(publicKeyFile)

    def loadKey(self, kfile: str):
        if kfile is None:
            self.__vk = None
        else:
            with open(kfile) as f:
                self.__vk = ecdsa.VerifyingKey.from_pem(f.read())

    def login(self, username, password, database=b"defaultdb"):
        req = schema_pb2_grpc.schema__pb2.LoginRequest(user=bytes(
            username, encoding='utf-8'), password=bytes(
                password, encoding='utf-8'
        ))
        try:
            self.__login_response = schema_pb2_grpc.schema__pb2.LoginResponse = \
                self.__stub.Login(
                    req
                )
        except ValueError as e:
            raise Exception(
                "Attempted to login on termninated client, channel has been shutdown") from e

        self.__stub = self.set_token_header_interceptor(self.__login_response)
        # Select database, modifying stub function accordingly
        request = schema_pb2_grpc.schema__pb2.Database(databaseName=database)
        resp = self.__stub.UseDatabase(request)
        self.__stub = self.set_token_header_interceptor(resp)

        self.__rs.init("%s/%s".format(self.__url, database), self.__stub)
        return self.__login_response

    def shutdown(self):
        self.channel.close()
        self.channel = None
        self.intercept_channel.close
        self.intercept_channel = None
        self.__rs = None

    def set_token_header_interceptor(self, response):
        try:
            token = response.token
        except AttributeError:
            token = response.reply.token
        self.header_interceptor = \
            header_manipulator_client_interceptor.header_adder_interceptor(
                'authorization', "Bearer " + token
            )
        try:
            self.intercept_channel = grpc.intercept_channel(
                self.channel, self.header_interceptor)
        except ValueError as e:
            raise Exception(
                "Attempted to login on termninated client, channel has been shutdown") from e
        return schema_pb2_grpc.ImmuServiceStub(self.intercept_channel)

    @property
    def stub(self):
        return self.__stub

    def healthCheck(self):
        return healthcheck.call(self.__stub, self.__rs)

    def get(self, key: bytes):
        return get.call(self.__stub, self.__rs, key)

    def getValue(self, key: bytes):
        ret = get.call(self.__stub, self.__rs, key)
        if ret is None:
            return None
        return ret.value

    def set(self, key: bytes, value: bytes):
        return setValue.call(self.__stub, self.__rs, key, value)

    def safeGet(self, key: bytes):
        warnings.warn("Call to deprecated safeGet. Use verifiedGet instead",
                      category=DeprecationWarning,
                      stacklevel=2
                      )
        return verifiedGet.call(self.__stub, self.__rs, key, verifying_key=self.__vk)

    def verifiedGet(self, key: bytes):
        return verifiedGet.call(self.__stub, self.__rs, key, verifying_key=self.__vk)

    def verifiedGetAt(self, key: bytes, atTx: int):
        return verifiedGet.call(self.__stub, self.__rs, key, atTx, self.__vk)

    def safeSet(self, key: bytes, value: bytes):
        warnings.warn("Call to deprecated safeSet. Use verifiedSet instead",
                      category=DeprecationWarning,
                      stacklevel=2
                      )
        return verifiedSet.call(self.__stub, self.__rs, key, value)

    def verifiedSet(self, key: bytes, value: bytes):
        return verifiedSet.call(self.__stub, self.__rs, key, value, self.__vk)

    def getAllValues(self, keys: list):
        resp = batchGet.call(self.__stub, self.__rs, keys)
        return resp

    def getAll(self, keys: list):
        resp = batchGet.call(self.__stub, self.__rs, keys)
        return {key: value.value for key, value in resp.items()}

    def setAll(self, kv: dict):
        return batchSet.call(self.__stub, self.__rs, kv)

    def changePassword(self, user, newPassword, oldPassword):
        request = schema_pb2_grpc.schema__pb2.ChangePasswordRequest(
            user=bytes(user, encoding='utf-8'),
            newPassword=bytes(newPassword, encoding='utf-8'),
            oldPassword=bytes(oldPassword, encoding='utf-8')
        )
        return changePassword.call(self.__stub, self.__rs, request)

    def changePermission(self, action, user, database, permissions):
        return changePermission.call(self.__stub, self.__rs, action, user, database, permissions)

    def createUser(self, user, password, permission, database):
        request = schema_pb2_grpc.schema__pb2.CreateUserRequest(
            user=bytes(user, encoding='utf-8'),
            password=bytes(password, encoding='utf-8'),
            permission=permission,
            database=database
        )
        return createUser.call(self.__stub, self.__rs, request)

    def listUsers(self):
        return listUsers.call(self.__stub, None)

    def databaseList(self):
        dbs = databaseList.call(self.__stub, self.__rs, None)
        return [x.databaseName for x in dbs.dblist.databases]

    def databaseUse(self, dbName: bytes):
        request = schema_pb2_grpc.schema__pb2.Database(databaseName=dbName)
        resp = databaseUse.call(self.__stub, self.__rs, request)
        # modify header token accordingly
        self.__stub = self.set_token_header_interceptor(resp)
        self.__rs.init(dbName, self.__stub)
        return resp

    def databaseCreate(self, dbName: bytes):
        request = schema_pb2_grpc.schema__pb2.Database(databaseName=dbName)
        return databaseCreate.call(self.__stub, self.__rs, request)

    def currentState(self):
        return currentRoot.call(self.__stub, self.__rs, None)

    def history(self, key: bytes, offset: int, limit: int, sortorder: bool):
        return history.call(self.__stub, self.__rs, key, offset, limit, sortorder)

    def logout(self):
        self.__stub.Logout(google_dot_protobuf_dot_empty__pb2.Empty())
        self.__login_response = None

    def scan(self, key: bytes, prefix: bytes, desc: bool, limit: int, sinceTx: int = None):
        return scan.call(self.__stub, self.__rs, key, prefix, desc, limit, sinceTx)

    def setReference(self, referredkey: bytes, newkey:  bytes):
        return reference.call(self.__stub, self.__rs, referredkey, newkey)

    def verifiedSetReference(self, referredkey: bytes, newkey:  bytes):
        return verifiedreference.call(self.__stub, self.__rs, referredkey, newkey, verifying_key=self.__vk)

    def zAdd(self, zset: bytes, score: float, key: bytes, atTx: int = 0):
        return zadd.call(self.__stub, self.__rs, zset, score, key, atTx)

    def verifiedZAdd(self, zset: bytes, score: float, key: bytes, atTx: int = 0):
        return verifiedzadd.call(self.__stub, self.__rs, zset, score, key, atTx, self.__vk)

    def zScan(self, zset: bytes, seekKey: bytes, seekScore: float,
              seekAtTx: int, inclusive: bool, limit: int, desc: bool, minscore: float,
              maxscore: float, sinceTx=None, nowait=False):
        return zscan.call(self.__stub, self.__rs, zset, seekKey, seekScore,
                          seekAtTx, inclusive, limit, desc, minscore,
                          maxscore, sinceTx, nowait)

    def txById(self, tx: int):
        return txbyid.call(self.__stub, self.__rs, tx)

    def verifiedTxById(self, tx: int):
        return verifiedtxbyid.call(self.__stub, self.__rs, tx, self.__vk)

    def sqlExec(self, stmt, params={}, noWait=False):
        """Executes an SQL statement
        Args:
            stmt: a statement in immudb SQL dialect.
            params: a dictionary of parameters to replace in the statement
            noWait: whether to wait for indexing. Set to True for fast inserts.

        Returns:
            An object with two lists: ctxs and dtxs, including transaction
            metadata for both the catalog and the data store.

            Each element of both lists contains an object with the Transaction ID
            (id), timestamp (ts), and number of entries (nentries).
        """

        return sqlexec.call(self.__stub, self.__rs, stmt, params, noWait)

    def sqlQuery(self, query, params={}):
        """Queries the database using SQL
        Args:
            query: a query in immudb SQL dialect.
            params: a dictionary of parameters to replace in the query

        Returns:
            A list of table names. For example:

            ['table1', 'table2']
        """
        return sqlquery.call(self.__stub, self.__rs, query, params)

    def listTables(self):
        """List all tables in the current database

        Returns:
            A list of table names. For example:

            ['table1', 'table2']
        """
        return listtables.call(self.__stub, self.__rs)

    def execAll(self, ops: list, noWait=False):
        return execAll.call(self.__stub, self.__rs, ops, noWait)

    def compactIndex(self):
        self.__stub.CompactIndex(google_dot_protobuf_dot_empty__pb2.Empty())

    def describeTable(self, table):
        return sqldescribe.call(self.__stub, self.__rs, table)



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, key: bytes, prefix: bytes, desc: bool, limit: int, sinceTx: int):
    if sinceTx == None:
        state = rs.get()
        sinceTx = state.txId
    request = schema_pb2_grpc.schema__pb2.ScanRequest(
        seekKey=key,
        prefix=prefix,
        desc=desc,
        limit=limit,
        sinceTx=sinceTx,
        noWait=False
    )
    msg = service.Scan(request)
    ret = {}
    for i in msg.entries:
        ret[i.key] = i.value
    return ret



#!/usr/bin/env python3
# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.constants import *
import hashlib
import struct
import copy
import immudb.htree
SHA256LEN = 32


class printable:
    def __repr__(self) -> str:
        ret = "class {}\n".format(type(self).__name__)
        for k in self.__dict__:
            v = self.__dict__[k]
            if type(v) == bytes:
                ret += "\t{}: {}\n".format(k, list(v))
            elif hasattr(v, '__getitem__') and type(v[0]) == bytes:
                ret += "\t{}:\n".format(k)
                for t in v:
                    ret += "\t- {}\n".format(list(t))
            else:
                ret += "\t{}: {}\n".format(k, v)
        return ret


class TXe(printable):
    def __init__(self):
        self.keyLen = None
        self.key = None
        self.ValueLen = None
        self.hValue = None
        self.vOff = None

    def SetKey(self, key: bytes):
        self.key = copy.copy(key)
        self.keyLen = len(key)

    def Digest(self):
        b = self.key+self.hValue
        return hashlib.sha256(b).digest()


class Tx(printable):
    def __init__(self):
        self.ID = None
        self.Ts = None
        self.BlTxID = None
        self.BlRoot = None
        self.PrevAlh = None
        self.nentries = None
        self.entries = None
        self.htree = None
        self.Alh = None
        self.InnerHash = None

    def BuildHashTree(self):
        digests = []
        for e in self.entries:
            digests.append(e.Digest())
        self.htree.BuildWith(digests)

    def CalcAlh(self):
        self.calcInnerhash()
        bi = struct.pack(">Q", self.ID)+self.PrevAlh+self.InnerHash
        self.Alh = hashlib.sha256(bi).digest()

    def calcInnerhash(self):
        bj = struct.pack(">QL", self.Ts, self.nentries)+self.eh()
        bj += struct.pack(">Q", self.BlTxID)+self.BlRoot
        self.InnerHash = hashlib.sha256(bj).digest()

    def eh(self):
        return self.htree.root

    def Proof(self, key: bytes):
        kindex = None
        # find index of element holding given key
        #kindex=next(k for k,v in enumerate(self.entries) if v.key==key)
        for k, v in enumerate(self.entries):
            if v.key == key:
                kindex = k
                break
        if kindex == None:
            raise LookupError
        return self.htree.InclusionProof(kindex)


def NewTxWithEntries(entries: list) -> Tx:
    tx = Tx()
    tx.ID = 0
    tx.entries = entries
    tx.nentries = len(entries)
    tx.htree = immudb.htree.HTree(len(entries))
    return tx


def TxFrom(stx) -> Tx:
    entries = []
    for e in stx.entries:
        i = TXe()
        i.hValue = DigestFrom(e.hValue)
        i.vOff = e.vOff
        i.ValueLen = int(e.vLen)
        i.SetKey(e.key)
        entries.append(i)
    tx = NewTxWithEntries(entries)
    tx.ID = stx.metadata.id
    tx.PrevAlh = DigestFrom(stx.metadata.prevAlh)
    tx.Ts = stx.metadata.ts
    tx.BlTxID = stx.metadata.blTxId
    tx.BlRoot = DigestFrom(stx.metadata.blRoot)
    tx.BuildHashTree()
    tx.CalcAlh()
    return tx


class TxMetadata(printable):
    def __init__(self):
        self.iD = None
        self.prevAlh = None
        self.ts = None
        self.nEntries = None
        self.eh = None
        self.blTxID = None
        self.blRoot = None

    def alh(self):
        bi = struct.pack(">Q", self.iD)+self.prevAlh
        bj = struct.pack(">QL", int(self.ts), int(self.nEntries))
        bj = bj+self.eh+struct.pack(">Q", self.blTxID)+self.blRoot
        innerHash = hashlib.sha256(bj).digest()
        bi = bi+innerHash
        return hashlib.sha256(bi).digest()


def TxMetadataFrom(txmFrom):
    txm = TxMetadata()
    txm.iD = txmFrom.id
    txm.prevAlh = DigestFrom(txmFrom.prevAlh)
    txm.ts = txmFrom.ts
    txm.nEntries = int(txmFrom.nentries)
    txm.eh = DigestFrom(txmFrom.eH)
    txm.blTxID = txmFrom.blTxId
    txm.blRoot = DigestFrom(txmFrom.blRoot)
    return txm


class KV(printable):
    def __init__(self, key: bytes, value: bytes):
        self.key = key
        self.value = value

    def Digest(self):
        if self.value == None:
            valdigest = hashlib.sha256(b'').digest()
        else:
            valdigest = hashlib.sha256(self.value).digest()
        return hashlib.sha256(self.key+valdigest).digest()


def EncodeKey(key: bytes):
    return SET_KEY_PREFIX+key


def EncodeKV(key: bytes, value: bytes):
    return KV(SET_KEY_PREFIX+key, PLAIN_VALUE_PREFIX+value)


def EncodeReference(key: bytes, referencedKey: bytes, atTx: int):
    refVal = REFERENCE_VALUE_PREFIX + \
        struct.pack(">Q", atTx)+SET_KEY_PREFIX+referencedKey
    return KV(SET_KEY_PREFIX+key, refVal)


def EncodeZAdd(zset: bytes, score: float, key: bytes, attx: int):
    ekey = SET_KEY_PREFIX+key
    zkey = SORTED_KEY_PREFIX
    zkey += struct.pack(">Q", len(zset))+zset
    zkey += struct.pack(">d", score)
    zkey += struct.pack(">Q", len(ekey))+ekey
    zkey += struct.pack(">Q", attx)
    return KV(zkey, None)


class LinearProof(printable):
    def __init__(self, sourceTxID: int, targetTxID: int, terms: list):
        self.sourceTxID = sourceTxID
        self.targetTxID = targetTxID
        self.terms = terms


def LinearProofFrom(lp) -> LinearProof:
    return LinearProof(lp.sourceTxId, lp.TargetTxId, lp.terms)


def DigestFrom(slicedDigest: bytes) -> bytes:
    d = copy.copy(slicedDigest[:SHA256LEN])
    return d


def DigestsFrom(slicedTerms: list) -> list:
    d = [copy.copy(i) for i in slicedTerms]
    return d


def VerifyInclusion(proof, digest: bytes, root) -> bool:
    if proof == None:
        return False
    leaf = LEAF_PREFIX+digest
    calcRoot = hashlib.sha256(leaf).digest()
    i = proof.leaf
    r = proof.width-1
    for t in proof.terms:
        b = NODE_PREFIX
        if i % 2 == 0 and i != r:
            b = b+calcRoot+t
        else:
            b = b+t+calcRoot
        calcRoot = hashlib.sha256(b).digest()
        i = i//2
        r = r//2
    return i == r and root == calcRoot


def leafFor(d: bytes) -> bytes:
    b = LEAF_PREFIX+d
    return hashlib.sha256(b).digest()


def VerifyDualProof(proof, sourceTxID, targetTxID, sourceAlh, targetAlh):
    if (proof == None or
        proof.sourceTxMetadata == None or
        proof.targetTxMetadata == None or
        proof.sourceTxMetadata.iD != sourceTxID or
            proof.targetTxMetadata.iD != targetTxID):
        return False
    if proof.sourceTxMetadata.iD == 0 or proof.sourceTxMetadata.iD > proof.targetTxMetadata.iD:
        return False
    if sourceAlh != proof.sourceTxMetadata.alh():
        return False
    if targetAlh != proof.targetTxMetadata.alh():
        return False
    if sourceTxID < proof.targetTxMetadata.blTxID and VerifyInclusionAHT(
            proof.inclusionProof,
            sourceTxID,
            proof.targetTxMetadata.blTxID,
            leafFor(sourceAlh),
            proof.targetTxMetadata.blRoot) == False:
        return False
    if proof.sourceTxMetadata.blTxID > 0 and VerifyConsistency(
            proof.consistencyProof,
            proof.sourceTxMetadata.blTxID,
            proof.targetTxMetadata.blTxID,
            proof.sourceTxMetadata.blRoot,
            proof.targetTxMetadata.blRoot) == False:
        return False
    if proof.targetTxMetadata.blTxID > 0 and VerifyLastInclusion(
            proof.lastInclusionProof,
            proof.targetTxMetadata.blTxID,
            leafFor(proof.targetBlTxAlh),
            proof.targetTxMetadata.blRoot) == False:
        return False
    if sourceTxID < proof.targetTxMetadata.blTxID:
        ret = VerifyLinearProof(
            proof.linearProof, proof.targetTxMetadata.blTxID, targetTxID, proof.targetBlTxAlh, targetAlh)
    else:
        ret = VerifyLinearProof(
            proof.linearProof, sourceTxID, targetTxID, sourceAlh, targetAlh)
    return ret


def VerifyInclusionAHT(iproof: list, i: int, j: int, iLeaf: bytes, jRoot: bytes) -> bool:
    if i > j or i == 0 or i < j and len(iproof) == 0:
        return False
    i1 = i - 1
    j1 = j - 1
    ciRoot = iLeaf
    for h in iproof:
        if i1 % 2 == 0 and i1 != j1:
            b = NODE_PREFIX+ciRoot+h
        else:
            b = NODE_PREFIX+h+ciRoot
        ciRoot = hashlib.sha256(b).digest()
        i1 = i1 >> 1
        j1 = j1 >> 1
    return jRoot == ciRoot


def VerifyConsistency(cproof: list, i: int, j: int, iRoot: bytes, jRoot: bytes) -> bool:
    if i > j or i == 0 or (i < j and len(cproof) == 0):
        return False
    if i == j and len(cproof) == 0:
        return iRoot == jRoot

    fn = i - 1
    sn = j - 1
    while fn % 2 == 1:
        fn = fn >> 1
        sn = sn >> 1
    ciRoot, cjRoot = cproof[0], cproof[0]
    for h in cproof[1:]:
        if fn % 2 == 1 or fn == sn:
            b = NODE_PREFIX+h+ciRoot
            ciRoot = hashlib.sha256(b).digest()
            b = NODE_PREFIX+h+cjRoot
            cjRoot = hashlib.sha256(b).digest()
            while fn % 2 == 0 and fn != 0:
                fn = fn >> 1
                sn = sn >> 1
        else:
            b = NODE_PREFIX+cjRoot+h
            cjRoot = hashlib.sha256(b).digest()
        fn = fn >> 1
        sn = sn >> 1
    return iRoot == ciRoot and jRoot == cjRoot


def VerifyLastInclusion(iproof: list, i: int, leaf: bytes, root: bytes) -> bool:
    if i == 0:
        return False
    i1 = i - 1
    iroot = leaf
    for h in iproof:
        b = NODE_PREFIX+h+iroot
        iroot = hashlib.sha256(b).digest()
        i1 >>= 1
    return root == iroot


def VerifyLinearProof(proof, sourceTxID: int, targetTxID: int, sourceAlh: bytes, targetAlh: bytes) -> bool:
    if proof == None or proof.sourceTxID != sourceTxID or proof.targetTxID != targetTxID:
        return False

    if (proof.sourceTxID == 0 or proof.sourceTxID > proof.targetTxID or
            len(proof.terms) == 0 or sourceAlh != proof.terms[0]):
        return False

    calculatedAlh = proof.terms[0]
    for i in range(1, len(proof.terms)):
        bs = struct.pack(">Q", proof.sourceTxID+i)+calculatedAlh+proof.terms[i]
        calculatedAlh = hashlib.sha256(bs).digest()

    return targetAlh == calculatedAlh



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from immudb.grpc import schema_pb2, schema_pb2_grpc
from immudb.rootService import RootService
from google.protobuf.empty_pb2 import Empty


@dataclass
class changePermissionResponse:
    reply: Empty


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, action, user, database, permissions):
    # Parameter names as of go-client.
    request = schema_pb2.ChangePermissionRequest(
        action=action, username=user, database=database, permission=permissions)
    msg = service.ChangePermission(request)
    return changePermissionResponse(
        reply=msg
    )



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

LEAF_PREFIX = b'\x00'
NODE_PREFIX = b'\x01'
ROOT_CACHE_PATH = ".immudbRoot"

PERMISSION_SYS_ADMIN = 255
PERMISSION_ADMIN = 254
PERMISSION_NONE = 0
PERMISSION_R = 1
PERMISSION_RW = 2

SET_KEY_PREFIX = b'\x00'
SORTED_KEY_PREFIX = b'\x01'

PLAIN_VALUE_PREFIX = b'\x00'
REFERENCE_VALUE_PREFIX = b'\x01'

OLDEST_FIRST = False
NEWEST_FIRST = True



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
import struct


@dataclass
class SetResponse:
    id: int
    verified: bool


@dataclass
class SafeGetResponse:
    id: int
    key: bytes
    value: bytes
    timestamp: int
    verified: bool
    refkey: bytes


@dataclass
class historyResponseItem:
    key: bytes
    value: bytes
    tx: int


@dataclass
class GetResponse:
    tx: int
    key: bytes
    value: bytes


@dataclass
class KeyValue():
    key: bytes
    value: bytes


@dataclass
class ZAddRequest():
    set: bytes
    score: float
    key: bytes
    atTx: int
    boundRef: bool
    noWait: bool

    def __init__(self, set, score, key, atTx=0, noWait=False):
        self.set = set
        self.score = score
        self.key = key
        self.atTx = atTx
        self.boundRef = atTx > 0
        self.noWait = noWait


@dataclass
class ReferenceRequest():
    key: bytes
    referencedKey: bytes
    atTx: int
    boundRef: bool
    noWait: bool

    def __init__(self, key, referencedKey, atTx=0, noWait=False):
        self.key = key
        self.referencedKey = referencedKey
        self.atTx = atTx
        self.boundRef = atTx > 0
        self.noWait = noWait


@dataclass
class ColumnDescription:
    name: str
    type: str
    nullable: bool
    index: str
    autoincrement: bool
    unique: bool



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService, State
from immudb.exceptions import VerificationException
from immudb import datatypes
import immudb.store


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, zset: bytes, score: float, key: bytes, atTx: int = 0, verifying_key=None):
    state = rs.get()
    request = schema_pb2.VerifiableZAddRequest(
        zAddRequest=schema_pb2.ZAddRequest(
            set=zset,
            score=score,
            key=key,
            atTx=atTx,
        ),
        proveSinceTx=state.txId
    )
    vtx = service.VerifiableZAdd(request)
    if vtx.tx.metadata.nentries != 1:
        raise VerificationException
    tx = immudb.store.TxFrom(vtx.tx)
    ekv = immudb.store.EncodeZAdd(zset, score, key, atTx)
    inclusionProof = tx.Proof(ekv.key)
    verifies = immudb.store.VerifyInclusion(
        inclusionProof, ekv.Digest(), tx.eh())
    if not verifies:
        raise VerificationException
    if tx.eh() != immudb.store.DigestFrom(vtx.dualProof.targetTxMetadata.eH):
        raise VerificationException
    if state.txId == 0:
        sourceID = tx.ID
        sourceAlh = tx.Alh
    else:
        sourceID = state.txId
        sourceAlh = immudb.store.DigestFrom(state.txHash)
    targetID = tx.ID
    targetAlh = tx.Alh
    verifies = immudb.store.VerifyDualProof(
        immudb.htree.DualProofFrom(vtx.dualProof),
        sourceID,
        targetID,
        sourceAlh,
        targetAlh,
    )
    if not verifies:
        raise VerificationException

    newstate = State(
        db=state.db,
        txId=targetID,
        txHash=targetAlh,
        publicKey=vtx.signature.publicKey,
        signature=vtx.signature.signature,
    )
    if verifying_key != None:
        newstate.Verify(verifying_key)
    rs.set(newstate)
    return datatypes.SetResponse(
        id=vtx.tx.metadata.id,
        verified=True,
    )



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import time
from immudb.exceptions import VerificationException

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService, State
from immudb import constants, datatypes, store, htree

#import base64


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, key: bytes, value: bytes, verifying_key=None):
    state = rs.get()
    # print(base64.b64encode(state.SerializeToString()))
    kv = schema_pb2.KeyValue(key=key, value=value)
    rawRequest = schema_pb2.VerifiableSetRequest(
        setRequest=schema_pb2.SetRequest(KVs=[kv]),
        proveSinceTx=state.txId,
    )
    verifiableTx = service.VerifiableSet(rawRequest)
    # print(base64.b64encode(verifiableTx.SerializeToString()))
    tx = store.TxFrom(verifiableTx.tx)
    inclusionProof = tx.Proof(constants.SET_KEY_PREFIX+key)
    ekv = store.EncodeKV(key, value)
    verifies = store.VerifyInclusion(inclusionProof, ekv.Digest(), tx.eh())
    if not verifies:
        raise VerificationException
    if tx.eh() != store.DigestFrom(verifiableTx.dualProof.targetTxMetadata.eH):
        raise VerificationException
    if state.txId == 0:
        sourceID = tx.ID
        sourceAlh = tx.Alh
    else:
        sourceID = state.txId
        sourceAlh = store.DigestFrom(state.txHash)
    targetID = tx.ID
    targetAlh = tx.Alh

    verifies = store.VerifyDualProof(
        htree.DualProofFrom(verifiableTx.dualProof),
        sourceID,
        targetID,
        sourceAlh,
        targetAlh,
    )
    if not verifies:
        raise VerificationException
    newstate = State(
        db=state.db,
        txId=targetID,
        txHash=targetAlh,
        publicKey=verifiableTx.signature.publicKey,
        signature=verifiableTx.signature.signature,
    )
    if verifying_key != None:
        newstate.Verify(verifying_key)
    rs.set(newstate)
    return datatypes.SetResponse(
        id=verifiableTx.tx.metadata.id,
        verified=verifies,
    )



from immudb.grpc import schema_pb2
from pprint import pformat


def py_to_sqlvalue(value):
    sqlValue = None
    typ = type(value)
    if value is None:
        sqlValue = schema_pb2.SQLValue(null=True)
    elif typ is int:
        sqlValue = schema_pb2.SQLValue(n=value)
    elif typ is bool:
        sqlValue = schema_pb2.SQLValue(b=value)
    elif typ is str:
        sqlValue = schema_pb2.SQLValue(s=value)
    elif typ in (bytes, bytearray):
        sqlValue = schema_pb2.SQLValue(bs=value)
    else:
        raise TypeError("Type not supported: %s".format(
            value.__class__.__name__))
    return sqlValue


def sqlvalue_to_py(sqlValue):
    if sqlValue.HasField("n"):
        return sqlValue.n
    elif sqlValue.HasField("b"):
        return bool(sqlValue.b)
    elif sqlValue.HasField("bs"):
        return sqlValue.bs
    elif sqlValue.HasField("s"):
        return sqlValue.s
    elif sqlValue.HasField("null"):
        return None
    else:
        raise TypeError("Type not supported: %s", sqlValue.WhichOneof("value"))



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb import datatypes
from immudb.grpc import schema_pb2, schema_pb2_grpc
from immudb.rootService import RootService


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, ops: list, noWait: bool):
    request_ops = []
    for op in ops:
        if type(op) is datatypes.KeyValue:
            request_ops.append(
                schema_pb2.Op(
                    kv=schema_pb2.KeyValue(
                        key=op.key,
                        value=op.value
                    )
                )
            )

        elif type(op) is datatypes.ZAddRequest:
            request_ops.append(
                schema_pb2.Op(
                    zAdd=schema_pb2.ZAddRequest(
                        set=op.set,
                        score=op.score,
                        key=op.key,
                        atTx=op.atTx,
                        boundRef=op.boundRef,
                        noWait=op.noWait
                    )
                )
            )
        elif type(op) is datatypes.ReferenceRequest:
            request_ops.append(
                schema_pb2.Op(
                    ref=schema_pb2.ReferenceRequest(
                        key=op.key,
                        referencedKey=op.referencedKey,
                        atTx=op.atTx,
                        boundRef=op.boundRef,
                        noWait=op.noWait
                    )
                )
            )
        else:
            raise("unknown op for execAll")

    request = schema_pb2.ExecAllRequest(Operations=request_ops, noWait=noWait)
    msg = service.ExecAll(request)
    return msg



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService, State
from immudb import datatypes, htree, store, exceptions


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, tx: int, verifying_key=None):
    state = rs.get()
    request = schema_pb2.VerifiableTxRequest(
        tx=tx,
        proveSinceTx=state.txId
    )
    try:
        vtx = service.VerifiableTxById(request)
    except Exception as e:
        if hasattr(e, 'details') and e.details() == 'tx not found':
            return None
        raise e
    dualProof = htree.DualProofFrom(vtx.dualProof)
    if state.txId <= vtx.tx.metadata.id:
        sourceid = state.txId
        sourcealh = store.DigestFrom(state.txHash)
        targetid = vtx.tx.metadata.id
        targetalh = dualProof.targetTxMetadata.alh()
    else:
        sourceid = vtx.tx.metadata.id
        sourcealh = dualProof.sourceTxMetadata.alh()
        targetid = state.txId
        targetalh = store.DigestFrom(state.txHash)
    verifies = store.VerifyDualProof(
        dualProof,
        sourceid,
        targetid,
        sourcealh,
        targetalh)
    if not verifies:
        raise exceptions.VerificationException
    newstate = State(
        db=state.db,
        txId=targetid,
        txHash=targetalh,
        publicKey=vtx.signature.publicKey,
        signature=vtx.signature.signature,
    )
    if verifying_key != None:
        newstate.Verify(verifying_key)
    rs.set(newstate)

    ret = []
    for t in vtx.tx.entries:
        ret.append(t.key[1:])
    return ret



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService, State
from immudb import constants, htree, store, datatypes
from immudb.exceptions import VerificationException

import sys


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, requestkey: bytes, atTx: int = None, verifying_key=None):
    state = rs.get()
    if atTx == None:
        req = schema_pb2.VerifiableGetRequest(
            keyRequest=schema_pb2.KeyRequest(key=requestkey),
            proveSinceTx=state.txId
        )
    else:
        req = schema_pb2.VerifiableGetRequest(
            keyRequest=schema_pb2.KeyRequest(key=requestkey, atTx=atTx),
            proveSinceTx=state.txId
        )
    ventry = service.VerifiableGet(req)
    inclusionProof = htree.InclusionProofFrom(ventry.inclusionProof)
    dualProof = htree.DualProofFrom(ventry.verifiableTx.dualProof)

    if ventry.entry.referencedBy == None or ventry.entry.referencedBy.key == b'':
        vTx = ventry.entry.tx
        kv = store.EncodeKV(requestkey, ventry.entry.value)
    else:
        vTx = ventry.entry.referencedBy.tx
        kv = store.EncodeReference(
            ventry.entry.referencedBy.key, ventry.entry.key, ventry.entry.referencedBy.atTx)

    if state.txId <= vTx:
        eh = store.DigestFrom(
            ventry.verifiableTx.dualProof.targetTxMetadata.eH)
        sourceid = state.txId
        sourcealh = store.DigestFrom(state.txHash)
        targetid = vTx
        targetalh = dualProof.targetTxMetadata.alh()
    else:
        eh = store.DigestFrom(
            ventry.verifiableTx.dualProof.sourceTxMetadata.eH)
        sourceid = vTx
        sourcealh = dualProof.sourceTxMetadata.alh()
        targetid = state.txId
        targetalh = store.DigestFrom(state.txHash)

    verifies = store.VerifyInclusion(inclusionProof, kv.Digest(), eh)
    if not verifies:
        raise VerificationException
    verifies = store.VerifyDualProof(
        dualProof,
        sourceid,
        targetid,
        sourcealh,
        targetalh)
    if not verifies:
        raise VerificationException
    newstate = State(
        db=state.db,
        txId=targetid,
        txHash=targetalh,
        publicKey=ventry.verifiableTx.signature.publicKey,
        signature=ventry.verifiableTx.signature.signature,
    )
    if verifying_key != None:
        newstate.Verify(verifying_key)
    rs.set(newstate)
    if ventry.entry.referencedBy != None and ventry.entry.referencedBy.key != b'':
        refkey = ventry.entry.referencedBy.key
    else:
        refkey = None

    return datatypes.SafeGetResponse(
        id=vTx,
        key=ventry.entry.key,
        value=ventry.entry.value,
        timestamp=ventry.verifiableTx.tx.metadata.ts,
        verified=verifies,
        refkey=refkey
    )
